{{
**************************************************
* File jvmVpCore: VP code (runs in seperate COG) *
**************************************************
}}

{{
  Virtual Peripherals are implemented as background code, meaning the code executes
  independantly from the interpreted java code.
}}


OBJ
  def: "jvmDefines"


VAR
  long vpId

  
PUB start(timerAddr)
{{ startup pasm routine }}
  vpId := cognew(@entry,timerAddr)+1
  if vpId > 0
    repeat while long[timerAddr] <> 0  'wait until COG initialized


PUB stop
{{ stop pasm routine }}
  if vpId > 0
    cogstop(vpId-1)

  
CON

{PRI isr | t,i,type
{{ Run all VP code }}
{{ This code needs to be converted to pasm }}
{{ It is the pasm entry }}
  repeat
    t := cnt
    repeat i from 0 to MAX_NUM_VPS-1                  'executes every 8.68 usec
      vpBank := (i<<4)                                'set currently accessed bank
      type := vpRambank[vpBank + nmPeriphISRVec]      'nmPeriphISRVec holds VP type
      case type
        NONE:   vpNone
        TIMER:  vpTimer
        ADC:    vpAdc
        DAC:    vpDac
        PWM:    vpPwm
        UARTRX: vpUartRx
        UARTTX: vpUartTx
        other:
    waitcnt(t+694) '694.4 = 80MHz*8.68usec          
}

CON

{PRI vpNone
{{ optional: wait an average time to minimize jitter }}
}

CON
  'file stamp\core\Timer.java (offsets within a VP bank)

{PRI vpTimer
{{ increment 32bits timer in bank specified by vpBank }}
{{ This code needs to be converted to pasm }}
{{ Since only one timer at any time, we simply use a long instead of bank registers }}
  vpTimerCount += 1
}    

CON
  'file stamp\core\ADC.java (offsets within a VP bank) 
  
{PRI vpAdc | inpin,outpin,accu,count
{{ this code needs to be converted to pasm }}
  inpin := >| vpRambank[vpBank + nmInPin]
  dira[inpin]~
  outpin := >| vpRambank[vpBank + nmOutPin]
  dira[outpin]~~
  if ina[inpin] == 0
    outa[outpin]~~ 'make outpin high
    accu := vpRambank[vpBank + nmAdcAccu]
    accu++
    if accu == 256 'prevent overflow
      accu := 255
    vpRambank[vpBank + nmAdcAccu] := accu
  else
    outa[outpin]~ 'make outpin low
  count := vpRambank[vpBank + nmAdcCount]
  count := (count+1) // 256
  vpRambank[vpBank + nmAdcCount] := count
  if count == 0
    vpRambank[vpBank + nmAdcValue] := accu 'write result
    vpRambank[vpBank + nmAdcAccu] := 0     'clear accu
}

CON
  'file stamp\core\DAC.java (offsets within a VP bank) 

{PRI vpDac | accu,pin
{{ this needs to be converted to pasm }}
  accu := vpRambank[vpBank + nmDacAccu]
  accu += vpRambank[vpBank + nmDacValue]
  vpRambank[vpBank + nmDacAccu] := accu
  pin := vpRambank[vpBank + nmDacPin]
  if accu > 255
    outa[pin]~~ 'pin high
  else
    outa[pin]~  'pin low  
  dira[pin]~~   'pin output
}

CON
  'file stamp\core\PWM.java (offsets within a VP bank) 
    
{PRI vpPwm | accu,pin,state,count
{{ this needs to be converted to pasm }}
  pin := vpRambank[vpBank + nmPWMPin]
  state := vpRambank[vpBank + nmPWMState]
  count := vpRambank[vpBank + nmPWMCount]
  count--
  if count == 0
    count := vpRambank[vpBank + nmPWMhighTime + state]
    vpRambank[vpBank + nmPWMCount] := count
    state := 1 - state
    vpRambank[vpBank + nmPWMState] := state
    if state == 1
      outa[pin]~~ 'pin high
    else
      outa[pin]~ 'pin low
  dira[pin]~     'pin output
}

CON
  'file stamp\core\Uart.java (offsets within a VP bank) 

{PRI vpUartRx | inbit,rxbyte,count,divide,rxpin
  rxpin := >| vpRambank[vpBank + nmUartPin]
  case vpRambank[vpBank + nmUartPort] 
    1: rxpin += 8
    2: rxpin += 16
    3: rxpin += 24
  dira[rxpin]~ 'make pin input
  if ina[rxpin] <> 0 'true if input high
    inbit := $80
  else
    inbit := $00
  if vpRambank[vpBank + nmUartCount] == 0
    if NOT inbit
      vpRambank[vpBank + nmUartCount] := 9 'in case start ready 9 bits
    divide := vpRambank[vpBank + nmUartBaud] * 3/2 'start delay
    vpRambank[vpBank + nmUartDivide] := divide
  divide := vpRambank[vpBank + nmUartDivide]
  divide--
  vpRambank[vpBank + nmUartDivide] := divide
  if divide == 0
    count := vpRambank[vpBank + nmUartCount]
    count--
    vpRambank[vpBank + nmUartCount] := count    
    if count <> 0
      rxbyte := vpRambank[vpBank + nmUartDataL]
      rxbyte := (rxbyte >> 1) | inbit
      vpRambank[vpBank + nmUartDataL] := rxbyte
    else
      'set rxflag (how to handle this ??)
    divide := vpRambank[vpBank + nmUartBaud] 'bit delay
    vpRambank[vpBank + nmUartDivide] := divide
}

{PRI vpUartTx | txpin,divide,count,txbyte
  divide := vpRambank[vpBank + nmUartDivide]
  divide--
  vpRambank[vpBank + nmUartDivide] := divide
  if divide == 0
    divide := vpRambank[vpBank + nmUartBaud] 'bit delay
    vpRambank[vpBank + nmUartDivide] := divide
    count := vpRambank[vpBank + nmUartCount]
    if count <> 0
    txbyte := (vpRambank[vpBank + nmUartDataH] << 8) + vpRambank[vpBank + nmUartDataL]
    txbyte >>= 1
    vpRambank[vpBank + nmUartDataL] := txbyte
    vpRambank[vpBank + nmUartDataH] := txbyte >> 8
    count--
    vpRambank[vpBank + nmUartCount] := count
    txpin := >| vpRambank[vpBank + nmUartPin]
    case vpRambank[vpBank + nmUartPort] 
      1: txpin += 8
      2: txpin += 16
      3: txpin += 24
    if txbyte & $4000 == 0
      outa[txpin]~~ 'pin high
    else
      outa[txpin]~ 'pin low
    dira[txpin]~~ 'make pin output
}

DAT
                        org     0
entry
                        mov     asmTimerCountAddr, PAR              'get address of vpTimerCount 
                        rdlong  asmVpRambankArea, asmTimerCountAddr 'get value stored in vpTimerCount (= offset to VP rambank area for this COG)
                        add     asmVpRambankArea, asmTimerCountAddr 'add to address of vpTimerCount
                        mov     asmBankOffset, #0                   'clear bankOffset
                        wrlong  asmBankOffset, asmTimerCountAddr    'clear vpTimerCount
 
isrAsm                  'forever loop for VP code
                        mov     time, cnt               'record current time
                        add     time, period            'establish next period

isrAsm_nextSlot         'get VP type from next VP slot and execute appropiate VP code
                        mov     asmRegBase, asmVpRambankArea   'calculate regBase
                        add     asmRegBase, asmBankOffset
                        rdbyte  asmVpType, asmRegBase          'regBase + 0 holds vpType or 0 if VP slot free
                        add     asmVpType, #vpCoreJumpTable    'calculate address holding address to jump to
                        movs    isrAsm_setJump, asmVpType      'enable to read this address
                        nop                                    'necessary ??
isrAsm_setJump          movd    isrAsm_dojump, 0-0             'put it in place
                        nop                                    'necessary ??

{
                        cmp     asmVpType, #def#NONE wz
              if_e      call    #vpNoneAsm

                        cmp     asmVpType, #def#TIMER wz
              if_e      call    #vpTimerAsm

                        cmp     asmVpType, #def#ADC wz
              if_e      call    #vpAdcAsm

                        cmp     asmVpType, #def#DAC wz
              if_e      call    #vpDacAsm

                        cmp     asmVpType, #def#PWM wz
              if_e      call    #vpPwmAsm

                        cmp     asmVpType, #def#UARTRX wz
              if_e      call    #vpUartRxAsm

                        cmp     asmVpType, #def#UARTTX wz
              if_e      call    #vpUartTxAsm
}

isrAsm_dojump           jmp     0-0                    'jump to appropiate VP code
isrAsm_retFromVP                                       'all VP code returns here
                        'move to next VP slot
                        add     asmBankOffset, #16
                        cmp     asmBankOffset, #48 wz   'we do 3 VP slots in this COG
              if_e      jmp     @isrAsm_nextSlot
                        mov     asmBankOffset, #0

isrAsm_waitForTick      'wait for next 8.68 usec tick
                        waitcnt time, period
                        jmp     @isrAsm
                      

vpNoneAsm               'wait an average time to minmize jitter
                        mov     asmCount, cnt
                        add     asmCount, #100          'wait 100 cycles
                        waitcnt asmCount, #100
vpNoneAsm_ret           jmp     @isrAsm_retFromVP


vpTimerAsm              'increment free running 32bit timer
                        rdlong  asmCount, asmTimerCountAddr
                        add     asmCount, #1
                        wrlong  asmCount, asmTimerCountAddr
vpTimerAsm_ret          jmp     @isrAsm_retFromVP



vpAdcAsm                'count high input at inpin over 256 loops
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmAdcInMask
                        rdlong  asmInMask, asmRegAddr
                        andn    dira, asmInMask                'make inpin input
                        add     asmRegAddr, #def#nmAdcOutMask - def#nmAdcInMask
                        rdlong  asmOutMask, asmRegAddr
                        or      dira, asmOutMask               'make outpin output
                        mov     asmInput, ina
                        and     asmInput, asmInMask wz
              if_nz     andn    outa, asmOutMask        'make outpin low
              if_nz     jmp     @vpAdcAsm_updateCount
                        or      outa, asmOutMask        'make output high
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmAdcAccu
                        rdbyte  asmAccu, asmRegAddr          'get accu
                        add     asmAccu, #1               'accu += 1
                        cmp     asmAccu, #256 wz
              if_e      sub     asmAccu, #1               'prevent overflow
                        wrbyte  asmAccu, asmRegAddr          'set accu
vpAdcAsm_updateCount
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmAdcCount
                        rdbyte  asmCount, asmRegAddr      'increment loop count
                        add     asmCount, #1
                        wrbyte  asmCount, asmRegAddr
                        cmp     asmCount, #256 wz
              if_ne     jmp     @vpAdcAsm_ret                 'if 256 loops
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmAdcValue
                        wrbyte  asmAccu,asmRegAddr            '  adcValue := accu
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmAdcAccu    '  accu := 0
                        wrbyte  asmRegAddr,#0     
vpAdcAsm_ret            jmp     @isrAsm_retFromVP


vpDacAsm                'make outpin low or high based on accu overflow
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmDacOutMask
                        rdlong  asmOutPin, asmRegAddr
                        or      dira, asmOutPin               'make outpin output
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmDacAccu
                        rdbyte  asmAccu, asmRegAddr                      'get accu
                        add     asmRegAddr, #def#nmDacValue - def#nmDacAccu
                        rdbyte  asmValue, asmRegAddr                     'get dacValue
                        add     asmAccu, asmValue                        'accu += dacValue
                        sub     asmRegAddr, #def#nmDacValue - def#nmDacAccu
                        wrbyte  asmAccu, asmRegAddr                      'set accu
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmDacPin
                        rdlong  asmOutMask, asmRegAddr                   'get outpin
                        shr     asmAccu, #8 wz                        'did accu overflow (eg. > 255 ? )
              if_z      andn    outa, asmOutMask                   'no: output low
              if_nz     or      outa, asmOutMask                   'yes: output high
                        or      dira, asmOutMask                   'make pin output
vpDacAsm_ret            jmp     @isrAsm_retFromVP


vpPwmAsm
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmPWMOutMask
                        rdlong  asmOutMask, asmRegAddr
                        or      dira, asmOutMask              'make outpin output
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmPWMState
                        rdbyte  asmState, asmRegAddr          'get state
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmPWMCount
                        rdbyte  asmCount, asmRegAddr          'get count
                        sub     asmCount, #1
                        tjnz    asmCount, @vpPwmAsm_updateCount
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmPWMhighTime
                        add     asmRegAddr, asmState
                        rdbyte  asmCount, asmRegAddr          'get new count
                        xor     asmState, #1 wz
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmPWMState
                        wrbyte  asmState, asmRegAddr          'set state
              if_z      andn    outa, asmOutMask              'low output
              if_nz     or      outa, asmOutMask              'high output
vpPwmAsm_updateCount
                        mov     asmRegAddr, asmRegBase
                        add     asmRegAddr, #def#nmPWMCount
                        wrbyte  asmCount, asmRegAddr          'set count
vpPwmAsm_ret            jmp     @isrAsm_retFromVP


vpUartRxAsm
vpUartRxAsm_ret         jmp     @isrAsm_retFromVP

vpUartTxAsm
vpUartTxAsm_ret         jmp     @isrAsm_retFromVP


'variables to monitor 8.68 usec tick
time                    long    0
period                  long    694             '694 ticks at 80MHz equals 8.68 usec

'VP code global registers
asmVpType               long    0
asmTimerCountAddr       long    0               'address of 32bit free running timer
asmJumpAddr             long    0               'address to jump to
asmVpRambankArea        long    0               'start of vpRambank area used in this COG
asmBankOffset           long    0               'increments with 16
asmRegBase              long    0               'base for VP registers
asmRegAddr              long    0               'currently accessed VP register

'vpTimer,vpADC,vpDAC,vpPWM temporary registers
asmAccu                 long    0
asmCount                long    0
asmValue                long    0
asmInPin                long    0
asmOutPin               long    0
asmState                long    0
asmInMask               long    0
asmOutMask              long    0
asmInput                long    0

vpCoreJumpTable         long    (@vpNoneAsm   - @entry)/4 '0
                        long    (@vpTimerAsm  - @entry)/4 '1
                        long    (@vpAdcAsm    - @entry)/4 '2
                        long    (@vpDacAsm    - @entry)/4 '3
                        long    (@vpPwmAsm    - @entry)/4 '4
                        long    (@vpUartRxAsm - @entry)/4 '5
                        long    (@vpUartTxAsm - @entry)/4 '6

                        fit     $1F0
  